@@ -7,10 +7,10 @@ ptxd_colgen_generate_sections()
     # ignore the '-q' option to get a valid collection.
     #
     unset PTXDIST_QUIET
-    ptxd_make_log "print-PACKAGES-m" | gawk '
+    ptxd_make_log "print-PACKAGES-m" "print-CROSS_PACKAGES-m" "print-HOST_PACKAGES-m" | gawk '
 	BEGIN {
 		FS = "=\"|\"|=";
-		col_in     = "'"${PTX_KGEN_DIR}"'" "/collection/ptx_collection.in";
+		col_in     = "'"${PTX_KGEN_DIR}"'" "/generated/ptx_collection.in";
 	}
 
 	$1 ~ /^PTX_MAP_TO_package/ {
@@ -22,7 +22,7 @@ ptxd_colgen_generate_sections()
 
 	$1 ~ /^PTX_MAP_._DEP/ {
 		pkg = gensub(/PTX_MAP_._DEP_/, "", "g", $1);
-		deps[pkg] = $2;
+		deps[pkg] = deps[pkg] " " $2;
 
 		next;
 	}
@@ -46,16 +46,25 @@ ptxd_colgen_generate_sections()
 			pkg = sorted[i];
 			pkg_lc = module_pkgs[pkg];
 
+			if (pkg_lc ~ /^host-|cross-/)
+				prompt = ""
+			else
+				prompt = "\tprompt \"" pkg_lc "\" if COLLECTION_MANUAL\n"
+
 			printf \
 				"config " pkg "\n"\
 				"\t"	"bool\n"\
-				"\t"	"prompt \"" pkg_lc "\" if COLLECTION_MANUAL\n"\
+				prompt\
 				"\t"	"default COLLECTION_ALL\n"	> col_in;
 
+
 			m = split(deps[pkg], dep, " ");
+			asort(dep, sdep);
+			last = "";
 			for (j = 1; j <= m; j++) {
-				if (dep[j] in module_pkgs)
-					print "\tselect " dep[j]	> col_in;
+				if (sdep[j] in module_pkgs && sdep[j] != last )
+					print "\tselect " sdep[j]	> col_in;
+				last = sdep[j]
 			}
 
 			printf "\n"					> col_in;
