@@ -4,8 +4,6 @@
 #               2008, 2009, 2010 by Marc Kleine-Budde <mkl@pengutronix.de>
 #               2011 by George McCollister <george.mccollister@gmail.com>
 #
-# See CREDITS for details about who has contributed to this project.
-#
 # For further information about the PTXdist project and license conditions
 # see the README file.
 #
@@ -28,6 +26,7 @@ ptxd_make_xpkg_fixup() {
     if [ -n "${pkg_xpkg_fixup_to}" ]; then
 	echo -n "install_fixup:	@${pkg_xpkg_fixup_from}@ -> ${pkg_xpkg_fixup_to} ... "
 	sed -i -e "s,@$pkg_xpkg_fixup_from@,$pkg_xpkg_fixup_to,g" "${pkg_xpkg_control}" || return
+	sed -i -e "s,@$pkg_xpkg_fixup_from@,$pkg_xpkg_fixup_to,g" "${pkg_xpkg_dbg_control}" || return
     else
 	echo -n "install_fixup:	append '${pkg_xpkg_fixup_from}' ... "
 	echo "${pkg_xpkg_fixup_from}" >> "${pkg_xpkg_control}" || return
