@@ -102,7 +102,7 @@ ptxd_source_kconfig() {
 	local ret
 
 	set -a
-	source "${1}" 2> /dev/null
+	source "${1}"
 	ret=$?
 	set +a
 
@@ -150,192 +150,6 @@ ptxd_get_ptxconf() {
 export -f ptxd_get_ptxconf
 
 
-
-#
-# migrate a config file
-# look in PTX_MIGRATEDIR for a migration handler and call it
-#
-# $1	part identifier ("ptx", "platform", "collection", "board", "user")
-#
-ptxd_kconfig_migrate() {
-	local part="${1}"
-	local assistent="${PTX_MIGRATEDIR}/migrate_${part}"
-
-	if [ \! -f "${assistent}" ]; then
-		return 0
-	fi
-
-	cp -- ".config" ".config.old" || return
-	sed -f "${assistent}" ".config.old" > ".config"
-	retval=$?
-
-	if [ $retval -ne 0 ]; then
-		ptxd_dialog_msgbox "error: error occured during migration"
-		return ${retval}
-	fi
-
-	if ! diff -u ".config.old" ".config" >/dev/null; then
-		ptxd_dialog_msgbox "info: successfully migrated '${file_dotconfig}'"
-	fi
-
-	return ${retval}
-}
-
-
-
-#
-# $1	what kind of config ("oldconfig", "menuconfig", "dep")
-# $2	part identifier ("ptx", "platform", "collection", "board", "user")
-# $...	optional parameters
-#
-ptxd_kconfig() {
-	local config="${1}"
-	local part="${2}"
-	local copy_back="true"
-
-	ptxd_kgen "${part}" || ptxd_bailout "error in kgen"
-
-	local file_kconfig file_dotconfig
-
-	case "${part}" in
-	ptx)
-		if [ -e "${PTXDIST_WORKSPACE}/Kconfig" ]; then
-			file_kconfig="${PTXDIST_WORKSPACE}/Kconfig"
-		else
-			file_kconfig="config/Kconfig"
-		fi
-		file_dotconfig="${PTXDIST_PTXCONFIG}"
-		;;
-	platform)
-		if [ -e "${PTXDIST_WORKSPACE}/platforms/Kconfig" ]; then
-			file_kconfig="${PTXDIST_WORKSPACE}/platforms/Kconfig"
-		else
-			file_kconfig="${PTXDIST_TOPDIR}/platforms/Kconfig"
-		fi
-		file_dotconfig="${PTXDIST_PLATFORMCONFIG}"
-		;;
-	collection)
-		ptxd_dgen || ptxd_bailout "error in dgen"
-
-		#
-		# "PTXDIST_COLLECTIONCONFIG" would overwrite
-		# certain "m" packages with "y".
-		#
-		# but "menuconfig collection" works only on the
-		# "m" packages, so unset PTXDIST_COLLECTIONCONFIG
-		# here.
-		#
-		PTXDIST_COLLECTIONCONFIG="" ptxd_colgen || ptxd_bailout "error in colgen"
-
-		file_kconfig="${PTXDIST_TOPDIR}/config/collection/Kconfig"
-		file_dotconfig="${3}"
-		;;
-	board)
-		if [ -e "${PTXDIST_WORKSPACE}/boardsetup/Kconfig" ]; then
-			file_kconfig="${PTXDIST_WORKSPACE}/boardsetup/Kconfig"
-		else
-			file_kconfig="${PTXDIST_TOPDIR}/config/boardsetup/Kconfig"
-		fi
-		file_dotconfig="${PTXDIST_BOARDSETUP}"
-		;;
-	user)
-		file_kconfig="${PTXDIST_TOPDIR}/config/setup/Kconfig"
-		file_dotconfig="${PTXDIST_PTXRC}"
-		;;
-	*)
-		echo
-		echo "${PTXDIST_LOG_PROMPT}error: invalid use of '${FUNCNAME} ${@}'"
-		echo
-		exit 1
-		;;
-	esac
-
-	local tmpdir
-	tmpdir="$(mktemp -d "${PTXDIST_TEMPDIR}/kconfig.XXXXXX")" || ptxd_bailout "unable to create tmpdir"
-	pushd "${tmpdir}" > /dev/null
-
-	ln -sf "${PTXDIST_TOPDIR}/rules" &&
-	ln -sf "${PTXDIST_TOPDIR}/config" &&
-	ln -sf "${PTXDIST_TOPDIR}/platforms" &&
-	ln -sf "${PTXDIST_WORKSPACE}" workspace &&
-	ln -sf "${PTX_KGEN_DIR}/${part}" generated || return
-
-	if [ -e "${file_dotconfig}" ]; then
-		cp -- "${file_dotconfig}" ".config" || return
-	fi
-
-	local conf="${PTXDIST_TOPDIR}/scripts/kconfig/conf"
-	local mconf="${PTXDIST_TOPDIR}/scripts/kconfig/mconf"
-	local nconf="${PTXDIST_TOPDIR}/scripts/kconfig/nconf"
-
-	export \
-	    KCONFIG_NOTIMESTAMP="1" \
-	    PROJECT="ptxdist" \
-	    FULLVERSION="${PTXDIST_VERSION_FULL}"
-
-	case "${config}" in
-	menuconfig)
-		"${mconf}" "${file_kconfig}"
-		;;
-	nconfig)
-		"${nconf}" "${file_kconfig}"
-		;;
-	oldconfig)
-		#
-		# In silent mode, we cannot redirect input. So use
-		# oldconfig instead of silentoldconfig if somebody
-		# tries to automate us.
-		#
-		ptxd_kconfig_migrate "${part}" &&
-		if tty -s; then
-			"${conf}" --silentoldconfig "${file_kconfig}"
-		else
-			"${conf}" --oldconfig "${file_kconfig}"
-		fi
-		;;
-	allmodconfig)
-		"${conf}" --allmodconfig "${file_kconfig}"
-		;;
-	allyesconfig)
-		"${conf}" --allyesconfig "${file_kconfig}"
-		;;
-	allnoconfig)
-		"${conf}" --allnoconfig "${file_kconfig}"
-		;;
-	dep)
-		copy_back="false"
-		yes "" | "${conf}" --writedepend "${file_kconfig}" &&
-		cp -- ".config" "${PTXDIST_DGEN_DIR}/${part}config"
-		;;
-	*)
-		echo
-		echo "${PTXDIST_LOG_PROMPT}error: invalid use of '${FUNCNAME} ${@}'"
-		echo
-		exit 1
-		;;
-	esac
-
-	local retval=${?}
-	unset \
-	    KCONFIG_NOTIMESTAMP \
-	    PROJECT \
-	    FULLVERSION
-
-	if [ ${retval} -eq 0 -a "${copy_back}" = "true" ]; then
-		cp -- .config "${file_dotconfig}" || return
-		if [ -f .config.old ]; then
-			cp -- .config.old "$(readlink -f "${file_dotconfig}").old" || return
-		fi
-	fi
-
-	popd > /dev/null
-	rm -fr "${tmpdir}"
-
-	return $retval
-}
-export -f ptxd_kconfig
-
-
 #
 # call make,
 # source shell libraries wich are used in make
@@ -352,8 +166,9 @@ ptxd_make() {
 			source "${lib}" || ptxd_bailout "failed to source lib: ${lib}"
 		done
 	done
+	ptxd_make_setup_progress "${@}" &&
 	${PTX_NICE:+nice -n ${PTX_NICE}} "${PTXCONF_SETUP_HOST_MAKE}" \
-	    "${PTX_MAKE_ARGS[@]}" "${PTXDIST_PARALLELMFLAGS_EXTERN}" \
+	    "${PTX_MAKE_ARGS[@]}" ${PTXDIST_PARALLELMFLAGS_EXTERN} ${PTXDIST_LOADMFLAGS} \
 	    -f "${RULESDIR}/other/Toplevel.make" "${@}" || return
 }
 
@@ -377,6 +192,8 @@ ptxd_make_log() {(
 		# stdout and logfile
 		exec {logout}> >(tee -a "${PTX_LOGFILE}")
 	else
+		# no user input
+		exec < /dev/null
 		# logfile only
 		exec {logout}>> "${PTX_LOGFILE}"
 	fi
@@ -431,13 +248,14 @@ ptxd_filter_dir() {
 
 	mkdir -p "${dstdir}" &&
 
-	tar -C "${srcdir}" -c . \
+	tar -c -C "${srcdir}" \
 		--exclude .svn \
 		--exclude .pc \
 		--exclude .git \
 		--exclude "*.in" \
 		--exclude "*.in.*" \
 		--exclude "*/*~" \
+		. \
 		| tar -C "${dstdir}" -x
 	check_pipe_status || return
 
@@ -489,6 +307,69 @@ ptxd_dumpstack() {
 	read
 }
 
+#
+# ptxd_replace_link - atomically replace a symlink
+# $1	source
+# $2	target
+#
+ptxd_replace_link() {
+	test -e "${2}" -a ! -L "${2}" && ptxd_bailout "'${2}' is not a link"
+
+	if [ "${1}" -ef "${2}" ]; then
+	    return
+	fi
+
+	ln -sfT "${1}" "${2}.tmp"
+	mv -T "${2}.tmp" "${2}"
+}
+export -f ptxd_replace_link
+
+#
+# ptxd_replace_copy_from_path
+#
+# atomically replace a file and use ptxd_in_path to find the source
+#
+# $1 variable name with paths separated by ":"
+# $2 filename to find within these paths
+# $$ the target file name
+#
+ptxd_replace_copy_from_path() {
+	if [ -d "${3}" ]; then
+	    ptxd_bailout "ptxd_replace_copy_from_path: '${3}' must not be a directory"
+	fi
+	if ! ptxd_in_path "${1}" "${2}"; then
+		ptxd_bailout "ptxd_replace_copy_from_path: could not find ${2} in ${1}"
+	fi
+	if [ -d "${ptxd_reply}" ]; then
+	    ptxd_bailout "ptxd_replace_copy_from_path: '${ptxd_reply}' must not be a directory"
+	fi
+	cp -a "${ptxd_reply}" "${3}.tmp" &&
+	mv "${3}.tmp" "${3}"
+}
+export -f ptxd_replace_copy_from_path
+
+ptxd_get_alternative_list() {
+    local prefix="${1%/}"
+    local file="${2#/}"
+    local -a layers
+    [ -n "${prefix}" -a -n "${file}" ] || return
+
+    ptxd_in_path PTXDIST_PATH_LAYERS
+    layers=( "${ptxd_reply[@]}" )
+
+    ptxd_reply=()
+    for layer in "${layers[@]}"; do
+	ptxd_reply=( \
+	    "${ptxd_reply[@]}" \
+	    "${layer}/${prefix}${PTXDIST_PLATFORMSUFFIX}/${file}" \
+	    "${layer}/${prefix}/${file}${PTXDIST_PLATFORMSUFFIX}" \
+	    "${layer}/${PTXDIST_PLATFORMCONFIG_SUBDIR}/${prefix}/${file}${PTXDIST_PLATFORMSUFFIX}" \
+	    "${layer}/${prefix}/${file}" \
+	    "${layer}/${PTXDIST_PLATFORMCONFIG_SUBDIR}/${prefix}/${file}" \
+	    )
+    done
+}
+export -f ptxd_get_alternative_list
 
 #
 # ptxd_get_alternative - look for files in platform, BSP and ptxdist
@@ -503,20 +384,8 @@ ptxd_dumpstack() {
 # array "ptxd_reply" containing the found files
 #
 ptxd_get_alternative() {
-    local prefix="${1%/}"
-    local file="${2}"
-    [ -n "${prefix}" -a -n "${file}" ] || return
-
-    list=( \
-	"${PTXDIST_WORKSPACE}/${prefix}${PTXDIST_PLATFORMSUFFIX}/${file}" \
-	"${PTXDIST_WORKSPACE}/${prefix}/${file}${PTXDIST_PLATFORMSUFFIX}" \
-	"${PTXDIST_PLATFORMCONFIGDIR}/${prefix}/${file}${PTXDIST_PLATFORMSUFFIX}" \
-	"${PTXDIST_WORKSPACE}/${prefix}/${file}" \
-	"${PTXDIST_PLATFORMCONFIGDIR}/${prefix}/${file}" \
-	"${PTXDIST_TOPDIR}/${prefix}/${file}" \
-	)
-
-    ptxd_get_path "${list[@]}"
+    ptxd_get_alternative_list "${@}" &&
+    ptxd_get_path "${ptxd_reply[@]}"
 }
 export -f ptxd_get_alternative
 
@@ -532,15 +401,48 @@ export -f ptxd_get_alternative
 ptxd_get_path() {
     [ -n "${1}" ] || return
 
-    ptxd_reply=( $(eval command ls -f -d "${@}" 2>/dev/null) )
+    local orig_IFS="${IFS}"
+    IFS="
+"
+    ptxd_reply=( $(command ls -f -d "${@}" 2>/dev/null) )
+    IFS="${orig_IFS}"
 
     [ ${#ptxd_reply[@]} -ne 0 ]
 }
 export -f ptxd_get_path
 
+#
+# ptxd_get_path_fitered - look for files and/or dirs
+#
+# Like ptxd_get_path but skips paths that are symlinks to /dev/null
+#
+# return:
+# 0 if files/dirs are found
+# 1 if no files/dirs are found
+#
+# array "ptxd_reply" containing the found files/dirs
+#
+ptxd_get_path_filtered() {
+    ptxd_get_path "${@}" || return
+
+    set -- "${ptxd_reply[@]}"
+    ptxd_reply=()
+
+    while [ $# -gt 0 ]; do
+	if [ "$(readlink -f "${1}")" != /dev/null ]; then
+	    ptxd_reply[${#ptxd_reply[@]}]="${1}"
+	fi
+	shift
+    done
+    [ ${#ptxd_reply[@]} -ne 0 ]
+}
+export -f ptxd_get_path_filtered
+
 #
 # ptxd_in_path - look for files and/or dirs
 #
+# Note: the make implemenation in ptx/in-path must produce the same result.
+#
 # $1 variable name with paths separated by ":"
 # $2 filename to find within these paths
 #
@@ -552,15 +454,65 @@ export -f ptxd_get_path
 #
 ptxd_in_path() {
 	local orig_IFS="${IFS}"
+	local -a paths tmp
+	local path
+	local relative
 	IFS=:
-	local -a paths
-	paths=( ${!1} )
+	tmp=( ${!1} )
 	IFS="${orig_IFS}"
-	paths=( "${paths[@]/%//${2}}" )
+	for path in "${tmp[@]}"; do
+	    local search
+	    case "${path}" in
+	    /*)
+		paths=( "${paths[@]}" "${path}" )
+		;;
+	    *)
+		if [ -n "${relative}" ]; then
+		    ptxd_bailout "More than one relative path found in ${1}"
+		fi
+		relative=true
+		ptxd_in_path PTXDIST_PATH_LAYERS "${path}" || continue
+		paths=( "${paths[@]}" "${ptxd_reply[@]}" )
+		;;
+	    esac
+	done
+	paths=( "${paths[@]/%/${2:+/}${2}}" )
 	ptxd_get_path "${paths[@]}"
 }
 export -f ptxd_in_path
 
+#
+# ptxd_in_platformconfigdir - find paths in platformconfigdir
+#
+# Note: the make implemenation in ptx/in-platformconfigdir must produce the
+# same result.
+#
+# $1 filename to find
+#
+# If the filename is absolute then return the filename unchanged.
+# Otherwise search in platformconfigdir in all layers for the file and
+# return the first hit.
+# If no file is found return ${PTXDIST_PLATFORMCONFIGDIR}/${1}
+#
+# return:
+# The result is written to stdout.
+#
+ptxd_in_platformconfigdir() {
+    case "${1}" in
+    /*)
+	echo "${1}"
+	;;
+    *)
+	if ptxd_in_path PTXDIST_PATH_PLATFORMCONFIGDIR "${1}"; then
+	    echo "${ptxd_reply[0]}"
+	else
+	    echo "${PTXDIST_PLATFORMCONFIGDIR}/${1}"
+	fi
+	;;
+    esac
+}
+export -f ptxd_in_platformconfigdir
+
 #
 # convert a relative or absolute path into an absolute path
 #
@@ -627,6 +579,23 @@ ptxd_abs2rel() {
 }
 export -f ptxd_abs2rel
 
+#
+# Converts a file URL into an absolute path
+#
+ptxd_file_url_path() {
+    local url="${1//file:\/\//}"
+    url="${url//lndir:\/\//}"
+    if [[ ! "${url}" =~ ^/ ]]; then
+	    # relative to absolute path
+	    if ptxd_in_path PTXDIST_PATH_LAYERS "${url}"; then
+		url="${ptxd_reply}"
+	    else
+		url="${PTXDIST_WORKSPACE}/${url}"
+	    fi
+    fi
+    echo "${url}"
+}
+export -f ptxd_file_url_path
 
 #
 # prints a path but removes non interesting prefixes
@@ -638,7 +607,7 @@ ptxd_print_path() {
     fi
 
     local path out
-    for path in ${PTXDIST_PATH//:/ }; do
+    for path in ${PTXDIST_WORKSPACE} ${PTXDIST_TOPDIR}; do
 	path="${path%/*}/"
 	out="${1/#${path}}"
 	if [ "${out}" != "${1}" ]; then
@@ -662,8 +631,8 @@ ptxd_human_to_number() {
 		exit 1
 	fi
 
-	num=$(echo "$1" | sed 's/m$/*1024*1024/I')
-	num=$(echo "$num" | sed 's/k$/*1024/I')
+	num=$(sed 's/m$/*1024*1024/I' <<< "${1}")
+	num=$(sed 's/k$/*1024/I' <<< "${num}")
 
 	echo $((num))
 }
@@ -677,8 +646,7 @@ ptxd_name_to_NAME() {
 		echo "usage: ptxd_name_to_NAME <pkg-name>"
 		exit 1
 	fi
-	name="$(echo "${1}" | tr 'a-z-' 'A-Z_')"
-	echo "${name}"
+	tr 'a-z-' 'A-Z_' <<< "${1}"
 }
 export -f ptxd_name_to_NAME
 
@@ -709,17 +677,36 @@ ptxd_debug(){
 
 ptxd_debug "Debugging is enabled - Turn off with PTX_DEBUG=false"
 
+ptxd_bailout_impl() {
+	echo
+	while [ $# -gt 0 ]; do
+		if [[ $# -eq 1 && "$1" =~ ^[0-9]+$ ]]; then
+			break
+		fi
+		echo -e "${PTXDIST_LOG_PROMPT}error: ${1}"
+		shift
+	done
+	echo
+	return ${1:-1}
+}
+export -f ptxd_bailout_impl
+
 #
 # print out error message and exit with status 1
 #
-# $1: error message
-# $2: optional exit value (1 is default)
+# $*: error messages, one line per argument
+#
+# If the last argument is a number, then it is used as exit value
+# (1 is default)
 #
 # ${PTXDIST_LOG_PROMPT}: to be printed before message
 #
 ptxd_bailout() {
-	echo "${PTXDIST_LOG_PROMPT}error: $1" >&2
-	exit ${2:-1}
+	if [ -n "${PTXDIST_FD_STDERR}" -a -n "${PTXDIST_QUIET}" ]; then
+		ptxd_bailout_impl "${@}"  >&${PTXDIST_FD_STDERR}
+	fi
+	ptxd_bailout_impl "${@}" >&2
+	exit
 }
 export -f ptxd_bailout
 
@@ -749,7 +736,10 @@ export -f ptxd_pedantic
 # ${PTXDIST_LOG_PROMPT}: to be printed before message
 #
 ptxd_warning() {
-	echo "${PTXDIST_LOG_PROMPT}warning: $1" >&2
+	while [ $# -gt 0 ]; do
+		echo -e "${PTXDIST_LOG_PROMPT}warning: $1" >&2
+		shift
+	done
 }
 export -f ptxd_warning
 
@@ -759,7 +749,7 @@ export -f ptxd_warning
 #
 ptxd_verbose() {
 	if [ "${PTXDIST_VERBOSE}" == "1" ]; then
-		echo "${PTXDIST_LOG_PROMPT}""${@}" >&2
+		echo "${PTXDIST_LOG_PROMPT}""${@}"
 	elif [ -n "${PTXDIST_FD_LOGFILE}" ]; then
 		echo "${PTXDIST_LOG_PROMPT}""${@}" >&9
 	fi
@@ -776,6 +766,16 @@ ptxd_eval() {
 }
 export -f ptxd_eval
 
+#
+# execute the arguments
+#
+ptxd_exec() {
+	ptxd_verbose "executing:" "${@}
+"
+	"${@}"
+}
+export -f ptxd_exec
+
 #
 # check if a previously executed pipe returned an error
 #
@@ -812,7 +812,6 @@ ptxd_lib_sysroot() {
 }
 export -f ptxd_lib_sysroot
 
-
 #
 # split ipkg filename into it's parts
 #
